from django.core.management.base import BaseCommand
from django.core.management import call_command
from django.db import connections
from django.db.migrations.executor import MigrationExecutor
from django.contrib.auth.models import Permission, ContentType, Group
from django.conf import settings  # Import settings

class Command(BaseCommand):
    help = 'Check and apply necessary migrations and ensure roles, permissions, and groups are set up correctly'

    def handle(self, *args, **kwargs):
        if settings.DEBUG:
            self.stdout.write(self.style.NOTICE('Checking for pending migrations...'))

        connection = connections['default']
        executor = MigrationExecutor(connection)
        targets = executor.loader.graph.leaf_nodes()
        
        if executor.migration_plan(targets):
            if settings.DEBUG:
                self.stdout.write(self.style.WARNING('Pending migrations found. Applying migrations...'))
            call_command('migrate')
            if settings.DEBUG:
                self.stdout.write(self.style.SUCCESS('All migrations applied successfully.'))
        else:
            if settings.DEBUG:
                self.stdout.write(self.style.SUCCESS('No pending migrations found.'))

        if settings.DEBUG:
            self.stdout.write(self.style.NOTICE('Checking roles and permissions...'))

        # Ensure necessary permissions exist
        permissions = {
            "create_user": "Create User",
            "edit_user": "Edit User",
            "create_provider": "Create Provider",
            "manage_provider_keys": "Manage Provider Keys",
            "manage_analytics_tokens": "Manage Analytics Tokens",
            "create_user_consent": "Create User Consent",
            "request_user_data": "Request User Data"
        }
        
        content_type, created = ContentType.objects.get_or_create(app_label='users', model='customuser')
        
        permission_objects = {}
        for codename, name in permissions.items():
            permission, created = Permission.objects.get_or_create(
                codename=codename,
                content_type=content_type,
                defaults={'name': name}
            )
            permission_objects[codename] = permission
            if settings.DEBUG:
                if created:
                    self.stdout.write(self.style.SUCCESS(f'Permission created: {name}'))
                else:
                    self.stdout.write(self.style.NOTICE(f'Permission already exists: {name}'))
    
        # Ensure auth groups exist and assign permissions
        groups_permissions = {
            "polaris_administrator": ["create_user","edit_user","create_provider","change_provider","manage_provider_keys","manage_analytics_tokens","create_user_consent","request_user_data"],
            "provider_manager": ["manage_provider_keys","create_provider"],
            "analyst": ["manage_analytics_tokens"],
            "user": ["create_user_consent"]
        }
        
        for group_name, perms in groups_permissions.items():
            group, created = Group.objects.get_or_create(name=group_name)
            if settings.DEBUG:
                if created:
                    self.stdout.write(self.style.SUCCESS(f'Group created: {group_name}'))
                else:
                    self.stdout.write(self.style.NOTICE(f'Group already exists: {group_name}'))
            
            # Assign permissions to group
            for perm in perms:
                if perm in permission_objects:
                    group.permissions.add(permission_objects[perm])
                    if settings.DEBUG:
                        self.stdout.write(self.style.SUCCESS(f'Assigned permission {perm} to group {group_name}'))
        
        if settings.DEBUG:
            self.stdout.write(self.style.SUCCESS('Roles, permissions, and groups check completed.'))
